'Program: USF Tic-Tac-Toe
'Author: Alejandro Gutierrez-Munoz
'University of South Florida
'CIS 6930-13. Machine Learning
'Professor: Dr. Lawrence O. Hall 

Public Class board

    Private gint_Turn As Int16
    Private Weights(2) As Double
    Private Features(2) As Int16
    Private BoardSquares() As String = {" ", " ", " ", " ", " ", " ", " ", " ", " "}
    Private GameHistory As List(Of String())
    Private Vtrain As List(Of Double)
    Private ComputerWin, UserWin, Tie As Int16
    Private WinningCombination As Int16
    Private GamesPlayed As Int16
    Private BackgroundWorkerRunning As Boolean
    Private logfile As IO.StreamWriter

    Private LR As Double = 0.1

    Private Sub board_Load(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles MyBase.Load

        Dim i As Int16
        Dim r As New Random(System.DateTime.Now.Millisecond)

        GameHistory = New List(Of String())
        Vtrain = New List(Of Double)

        GamesPlayed = 0
        rdbUser.Checked = True
        gint_Turn = 1
        btnStop.Enabled = False
        numberGames.Enabled = False
        lblNoGames.Enabled = False

        'intialize weights at random
        For i = 0 To Weights.GetLength(0) - 1
            Weights(i) = r.Next(-1, 1) 'between -1 and 1
        Next

        'Create an empty log file
        logfile = New IO.StreamWriter(AppDomain.CurrentDomain.SetupInformation.ApplicationBase() & "/tictactoe_log.txt")

        'write header line
        logfile.WriteLine("Game,Outcome,ComputerWin,Tie,UserWin")

    End Sub

    Private Sub board_Disposed(ByVal sender As Object, ByVal e As System.EventArgs) Handles Me.Disposed
        logfile.Close()
    End Sub

    Private Sub sq_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles sq0.Click, sq1.Click, sq2.Click, sq3.Click, sq4.Click, sq5.Click, sq6.Click, sq7.Click, sq8.Click

        Dim SelectedSquare As Int16

        ' check if the square is already filled
        If (sender.text = " " And rdbUser.Checked) Then

            ' check who's turn it is
            If (gint_Turn = 1) Then
                sender.Text = "X"

                SelectedSquare = CInt(sender.Tag)

                'using the label's Tag save the move on the board
                BoardSquares(SelectedSquare) = sender.Text

                'Record game history
                GameHistory.Add(BoardSquares.Clone)

                If GameStillPlaying() Then
                    gint_Turn = 2 'change the turn to the opponent
                    SelectedSquare = SelectComputerMove()

                    'Update GUI
                    UpdateBoardGui(SelectedSquare, "O")

                    If GameStillPlaying() Then
                        gint_Turn = 1
                    End If

                End If

            End If

        End If

    End Sub

    Private Function GameStillPlaying() As Boolean


        Dim GameOver As Boolean
        Dim BoardResult As Int16
        Dim SqColor As Color
        Dim finalWinningCombination As Int16
        Dim logLine As String


        BoardResult = CheckBoardState(BoardSquares)
        finalWinningCombination = WinningCombination

        'check if current move finishes the game
        Select Case BoardResult

            Case 0
                GameOver = False
            Case 1
                UserWin += 1
                lblUserWin.Text = Strings.Right("   " & UserWin.ToString, 3)
                BoardTable.Enabled = False
                GameOver = True
                SqColor = Color.Blue
            Case 2
                ComputerWin += 1
                lblComputerWin.Text = Strings.Right("   " & ComputerWin.ToString, 3)
                BoardTable.Enabled = False
                SqColor = Color.Red
                GameOver = True
            Case 3
                Tie += 1
                lblTie.Text = Strings.Right("    " & Tie.ToString, 3)
                BoardTable.Enabled = False
                GameOver = True
        End Select

        'Game is over, Call the CRITIC system to determine <b,Vtrain(b)>
        ' Then call the GENERALIZER system to learn from current Vtrain(b) values
        If GameOver Then
            Critic()
            Generalizer()


            'Update GUI Only if the user is playing
            If rdbUser.Checked Then
                If (finalWinningCombination <> -1) Then
                    Select Case finalWinningCombination
                        Case 0
                            sq0.BackColor = SqColor
                            sq1.BackColor = SqColor
                            sq2.BackColor = SqColor
                        Case 1
                            sq3.BackColor = SqColor
                            sq4.BackColor = SqColor
                            sq5.BackColor = SqColor
                        Case 2
                            sq6.BackColor = SqColor
                            sq7.BackColor = SqColor
                            sq8.BackColor = SqColor
                        Case 3
                            sq0.BackColor = SqColor
                            sq3.BackColor = SqColor
                            sq6.BackColor = SqColor
                        Case 4
                            sq1.BackColor = SqColor
                            sq4.BackColor = SqColor
                            sq7.BackColor = SqColor
                        Case 5
                            sq2.BackColor = SqColor
                            sq5.BackColor = SqColor
                            sq8.BackColor = SqColor
                        Case 6
                            sq0.BackColor = SqColor
                            sq4.BackColor = SqColor
                            sq8.BackColor = SqColor
                        Case 7
                            sq2.BackColor = SqColor
                            sq4.BackColor = SqColor
                            sq6.BackColor = SqColor
                    End Select
                End If
            End If

            GamesPlayed += 1

            If chkLogFile.Checked Then

                logLine = GamesPlayed.ToString & ","

                'log tracks wins by computer
                Select Case BoardResult
                    Case 1
                        logLine &= "Lost,"
                    Case 2
                        logLine &= "Won,"
                    Case 3
                        logLine &= "Tied,"
                End Select

                logLine &= ComputerWin.ToString & "," & Tie.ToString & "," & UserWin.ToString

                logfile.WriteLine(logLine)
            End If

        End If

        Return Not GameOver

    End Function

    Private Function CheckBoardState(ByVal CurrentBoard() As String) As Int16

        'This function will return 1 if User won, 2 if Computer Won, 
        '3 if there is a tie, and 0 if game's still playing

        'Also, the features new values is calcuated based on the board being
        'passed as a parameter


        Dim Result As Int16
        Dim i As Int16
        Dim WinningCombinations(7) As String

        'First create all of the WinningCombinations based on the current board
        WinningCombinations(0) = CurrentBoard(0) & CurrentBoard(1) & CurrentBoard(2)
        WinningCombinations(1) = CurrentBoard(3) & CurrentBoard(4) & CurrentBoard(5)
        WinningCombinations(2) = CurrentBoard(6) & CurrentBoard(7) & CurrentBoard(8)
        WinningCombinations(3) = CurrentBoard(0) & CurrentBoard(3) & CurrentBoard(6)
        WinningCombinations(4) = CurrentBoard(1) & CurrentBoard(4) & CurrentBoard(7)
        WinningCombinations(5) = CurrentBoard(2) & CurrentBoard(5) & CurrentBoard(8)
        WinningCombinations(6) = CurrentBoard(0) & CurrentBoard(4) & CurrentBoard(8)
        WinningCombinations(7) = CurrentBoard(2) & CurrentBoard(4) & CurrentBoard(6)

        'Initialize Features for new counts based on current board
        For i = 0 To Features.GetLength(0) - 1
            Features(i) = 0
        Next

        ' Assume all spaces are filled, and game's a tie
        Result = 3

        'if there are any spaces, then game's still playing
        For i = 0 To CurrentBoard.GetLength(0) - 1
            If CurrentBoard(i) = " " Then
                Result = 0 ' Still playing
                i = CurrentBoard.GetLength(0) 'break for
            End If
        Next

        WinningCombination = -1

        'Check all possible winning combinations
        For i = 0 To WinningCombinations.GetLength(0) - 1

            Select Case WinningCombinations(i)
                Case "X  ", " X ", "  X"
                    If Features(1) < 1 Then
                        Features(1) = 1
                    End If
                Case "XX ", " XX", "X X"
                    If Features(1) < 2 Then
                        Features(1) = 2
                    End If
                Case "XXX"
                    Features(1) = 3
                    Result = 1 'Xs win
                    WinningCombination = i
                Case "O  ", " O ", "  O"
                    If Features(2) < 1 Then
                        Features(2) = 1
                    End If
                Case "OO ", " OO", "O O"
                    If Features(2) < 2 Then
                        Features(2) = 2
                    End If
                Case "OOO"
                    Features(2) = 3
                    Result = 2 'Os win
                    WinningCombination = i
            End Select

        Next


        Return Result

    End Function

    Private Function VirtualPlayer() As Int16

        Dim Playing As Boolean
        Dim r As New Random(System.DateTime.Now.Millisecond)
        Dim r2 As New Random(System.DateTime.Now.Millisecond + 1)
        Dim r3 As New Random(System.DateTime.Now.Millisecond + 3)
        Dim i As Int16

        Playing = True

        Do While (Playing)

            'search one empty square at random, use two random generator selected at random as well
            If r.Next(0, 1) Then
                i = r2.Next(0, 9)
            Else
                i = r3.Next(0, 9)
            End If

            If BoardSquares(i) = " " Then
                BoardSquares(i) = "X" ' Make play
                Playing = False
            End If

        Loop

        'Record game history
        GameHistory.Add(BoardSquares.Clone)

        'return the squared played
        Return i


    End Function


    Private Function SelectComputerMove() As Int16

        Dim V() As Double = {-9.999E+100, -9.999E+100, -9.999E+100, -9.999E+100, -9.999E+100, -9.999E+100, -9.999E+100, -9.999E+100, -9.999E+100}
        Dim BestBoard As Int16 = 0
        Dim BoardResult As Int16
        Dim MaxV As Double = -9.999E+100
        Dim i As Int16

        'Analyze V(b) for every possible valid move
        For i = 0 To BoardSquares.GetLength(0) - 1

            If BoardSquares(i) = " " Then
                BoardSquares(i) = "O" ' Make play
                BoardResult = CheckBoardState(BoardSquares) ' Check Board State
                V(i) = Vlearned() ' Calculate V(b) based on current features
                BoardSquares(i) = " " ' Reverse Computer Play
            End If

        Next

        'Select Play with highest V(b) value
        For i = 0 To V.GetLength(0) - 1

            If (V(i) > MaxV And BoardSquares(i) = " ") Then
                MaxV = V(i)
                BestBoard = i
            End If

        Next

        'make the play
        BoardSquares(BestBoard) = "O"

        'Record game history
        GameHistory.Add(BoardSquares.Clone)

        Return BestBoard


    End Function


    Private Sub UpdateBoardGui(ByVal BoardSquare As Int16, ByVal Symbol As String)

        'Update the GUI
        Select Case BoardSquare
            Case 0
                sq0.Text = Symbol
            Case 1
                sq1.Text = Symbol
            Case 2
                sq2.Text = Symbol
            Case 3
                sq3.Text = Symbol
            Case 4
                sq4.Text = Symbol
            Case 5
                sq5.Text = Symbol
            Case 6
                sq6.Text = Symbol
            Case 7
                sq7.Text = Symbol
            Case 8
                sq8.Text = Symbol
        End Select

    End Sub

    Private Function Vlearned() As Double

        Dim Result As Double = 0.0
        Dim i As Int16

        For i = 0 To Weights.GetLength(0) - 1
            Result += Weights(i) * Features(i)
        Next

        Return Result

    End Function


    Private Sub Critic()

        Dim i As Int16
        Dim BoardResult As Int16
        Dim Successor(8) As String

        'Based on game history, determine the value for the training examples <b,Vtrain(b)>
        For i = 0 To GameHistory.Count - 1

            If (i = GameHistory.Count - 1) Then
                Successor = GameHistory(GameHistory.Count - 1) 'final board of the game 

                Select Case CheckBoardState(Successor)
                    Case 1 ' User(Xs) wins
                        Vtrain.Add(-1.0)
                    Case 2 ' Computer(Os) winds
                        Vtrain.Add(1.0)
                    Case 3 ' Tie
                        Vtrain.Add(0.0)
                End Select


            Else
                'Successor is the board state following the program's move and the opponent's move
                Successor = GameHistory(i + 1)

                ' Vtrain(b) <- Vlearned(successor(b)
                BoardResult = CheckBoardState(Successor) 'CheckboardState sets the features vector to the succesor board
                Vtrain.Add(Vlearned())
            End If


        Next

    End Sub

    Private Sub Generalizer()

        Dim i As Int16
        Dim j As Int16
        Dim TrainError As Double
        Dim CurrentValue As Double
        Dim BoardResult As Int16
        Dim TrainingBoard(8) As String

        'Apply LMS to update the weights
        For i = 0 To GameHistory.Count - 1

            'Calculate V(b) based on current Vlearned function
            TrainingBoard = GameHistory(i)
            BoardResult = CheckBoardState(TrainingBoard)
            CurrentValue = Vlearned()

            'calculate the error
            TrainError = Vtrain(i) - CurrentValue

            For j = 0 To Weights.GetLength(0) - 1
                Weights(j) = Weights(j) + (LR * TrainError * Features(j))
            Next

        Next

    End Sub

    Private Sub ClearBoard()

        Dim i As Int16

        'Reset All squares
        sq0.Text = " "
        sq1.Text = " "
        sq2.Text = " "
        sq3.Text = " "
        sq4.Text = " "
        sq5.Text = " "
        sq6.Text = " "
        sq7.Text = " "
        sq8.Text = " "

        'Clear back color only if user is playing
        If rdbUser.Checked Then
            sq0.BackColor = Color.Transparent
            sq1.BackColor = Color.Transparent
            sq2.BackColor = Color.Transparent
            sq3.BackColor = Color.Transparent
            sq4.BackColor = Color.Transparent
            sq5.BackColor = Color.Transparent
            sq6.BackColor = Color.Transparent
            sq7.BackColor = Color.Transparent
            sq8.BackColor = Color.Transparent
        End If

        'clear boardsquares
        For i = 0 To BoardSquares.GetLength(0) - 1
            BoardSquares(i) = " "
        Next

        GameHistory.Clear()
        Vtrain.Clear()

        'Record game board 0, emtpy board
        GameHistory.Add(BoardSquares.Clone)

        gint_Turn = 1
        BoardTable.Enabled = True


    End Sub

    Private Sub bgWorker_DoWork(ByVal sender As Object, ByVal e As System.ComponentModel.DoWorkEventArgs) Handles bgWorker.DoWork

        Dim SelectedSquare As Int16

        BackgroundWorkerRunning = True

        'The background worker will run until the total number of games
        'has been reached, or a cancellation event has happened
        Do While (GamesPlayed < numberGames.Value) And BackgroundWorkerRunning

            If gint_Turn = 1 Then
                SelectedSquare = VirtualPlayer()
                bgWorker.ReportProgress(0, SelectedSquare)
            Else
                SelectedSquare = SelectComputerMove()
                bgWorker.ReportProgress(0, SelectedSquare)
            End If

            'this pause is intentional so we can remove the background worker
            'thread from the processor and allow the main program thread to be
            'serviced in case there is any external request.
            Threading.Thread.Sleep(3)

        Loop

    End Sub

    Private Sub bgWorker_ProgressChanged(ByVal sender As Object, ByVal e As System.ComponentModel.ProgressChangedEventArgs) Handles bgWorker.ProgressChanged


        If gint_Turn = 1 Then
            UpdateBoardGui(e.UserState, "X")
            gint_Turn = 2
        Else
            UpdateBoardGui(e.UserState, "O")
            gint_Turn = 1
        End If

        If Not GameStillPlaying() Then
            ClearBoard()
        End If

        If bgWorker.CancellationPending Then
            BackgroundWorkerRunning = False
        End If

    End Sub

    Private Sub bgWorker_RunWorkerCompleted(ByVal sender As Object, ByVal e As System.ComponentModel.RunWorkerCompletedEventArgs) Handles bgWorker.RunWorkerCompleted

        Dim percentageWon As Double

        ClearBoard()
        btnNewGame.Enabled = True
        btnStop.Enabled = False


        percentageWon = (ComputerWin / GamesPlayed) * 100.0
        MsgBox("Computer won " & Math.Round(percentageWon, 2) & "% of the games.", MsgBoxStyle.Information)

    End Sub


    Private Sub btnNewGame_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles btnNewGame.Click

        ClearBoard()

        If rdbVirtualUser.Checked Then
            btnNewGame.Enabled = False
            btnStop.Enabled = True
            bgWorker.RunWorkerAsync()
        End If

    End Sub


    Private Sub btnStop_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles btnStop.Click

        bgWorker.CancelAsync()
        btnNewGame.Enabled = True
        btnStop.Enabled = False

    End Sub

    Private Sub rdbUser_CheckedChanged(ByVal sender As Object, ByVal e As System.EventArgs) Handles rdbUser.CheckedChanged

        If rdbUser.Checked Then
            rdbVirtualUser.Checked = False
            btnNewGame.Text = "New Game"
            numberGames.Enabled = False
            lblNoGames.Enabled = False

        End If

    End Sub

    Private Sub rdbVirtualUser_CheckedChanged(ByVal sender As Object, ByVal e As System.EventArgs) Handles rdbVirtualUser.CheckedChanged

        If rdbVirtualUser.Checked Then
            rdbUser.Checked = False
            btnNewGame.Text = "Start"
            numberGames.Enabled = True
            lblNoGames.Enabled = True

        End If

    End Sub


End Class

